#include <stdio.h>
#include <stdlib.h> 
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#ifdef _WIN32
#include <io.h>
#else
#include <unistd.h>
#endif
#include "image.h"

#ifndef SEEK_SET
#  define SEEK_SET 0
#endif

#ifdef _WIN32
#define OPEN_FLAGS O_BINARY | O_RDONLY
#else
#define OPEN_FLAGS O_RDONLY
#endif

#define MIN_IMAGE	8
#define	MAX_IMAGE	2048


/******************************************************************************/
//
//	Definitionen der Funktionen
//
/******************************************************************************/
ImageRec *ReadTexture (const char *File) {
	ImageRec *Rec;
	if ((Rec = LoadBMP (File,FALSE)) == NULL)
		if ((Rec = LoadTGA (File,FALSE)) == NULL) 
			if ((Rec = LoadRGB (File,FALSE)) == NULL) 	
				printf ("Unsupported image format!\n");
	return Rec;
}

int GetFileSize (char *path) {
	struct stat file_stats;   
	if (stat (path,&file_stats))
    	return -1;   
	return file_stats.st_size;
}   

int GetFileSize (int FileDes) {
	struct stat file_stats;   
	if (fstat (FileDes,&file_stats))
    	return -1;   
	return file_stats.st_size;
}   

////////////////////////////////////////////////////////////////////////////////
//	RGB Funktionen
////////////////////////////////////////////////////////////////////////////////
static void ConvertShort (unsigned short *array, unsigned int length) {
    unsigned short b1, b2;
    unsigned char *ptr;

    ptr = (unsigned char *)array;
    while (length--) {
	b1 = *ptr++;
	b2 = *ptr++;
	*array++ = (b1 << 8) | (b2);
    }
}

static void ConvertUint (GLuint *array, unsigned int length) {
    unsigned int b1, b2, b3, b4;
    unsigned char *ptr;

    ptr = (unsigned char *)array;
    while (length--) {
	b1 = *ptr++;
	b2 = *ptr++;
	b3 = *ptr++;
	b4 = *ptr++;
	*array++ = (b1 << 24) | (b2 << 16) | (b3 << 8) | (b4);
    }
}

rawImageRec *RawImageOpen (const char *fileName,bool Verbose) {
    union {
		int testWord;
		char testByte[4];
    } endianTest;
	
    rawImageRec *raw;
    int swapFlag;
    int x;

    endianTest.testWord = 1;
    swapFlag = (endianTest.testByte[0] == 1);
    raw = (rawImageRec *)malloc (sizeof (rawImageRec));
	memset (raw,0,sizeof (rawImageRec));
    if (raw == NULL) {
		fprintf(stderr, "Out of memory!\n");
		exit(1);
	    }
    if ((raw->file = fopen (fileName, "rb")) == NULL) {
		if (Verbose)
			printf ("%s: Cannot open file for reading!\n",fileName);
		free (raw);
		return NULL;
	    }	
    if (fread (raw, 1, 12, raw->file) != 12) {
		if (Verbose)
			printf ("%s: Could not read file correctly!\n",fileName);
		fclose (raw -> file);
		free (raw);
		return NULL;
		}
    if (swapFlag) 
		ConvertShort (&raw->imagic, 6);

	if (raw -> sizeX > MAX_IMAGE || raw -> sizeX < MIN_IMAGE ||
		raw -> sizeY > MAX_IMAGE || raw -> sizeY < MIN_IMAGE) {
		fclose (raw -> file);
		free (raw);
		return NULL;
		}
    raw->tmp  = (unsigned char *)malloc(raw->sizeX*256);
    raw->tmpR = (unsigned char *)malloc(raw->sizeX*256);
    raw->tmpG = (unsigned char *)malloc(raw->sizeX*256);
    raw->tmpB = (unsigned char *)malloc(raw->sizeX*256);
    if (raw->tmp == NULL || raw->tmpR == NULL || raw->tmpG == NULL ||
		raw->tmpB == NULL) {
		fprintf(stderr, "Out of memory!\n");
		exit(1);
	    }
    if ((raw->type & 0xFF00) == 0x0100) {
		x = raw->sizeY * raw->sizeZ * (int) sizeof(GLuint);
		raw->rowStart = (GLuint *)malloc(x);
		raw->rowSize  = (GLint *)malloc(x);
		if (raw->rowStart == NULL || raw->rowSize == NULL) {
	    	fprintf(stderr, "Out of memory!\n");
		    exit(1);
			}
		raw->rleEnd = 512 + (2 * x);
		fseek (raw->file, 512, SEEK_SET);
		if (fread (raw->rowStart, 1, x, raw->file) != (UINT)x) {
			if (Verbose)
				printf ("%s: Could not read file correctly!\n",fileName);
			RawImageClose (raw);
			return NULL;
			}
		if (fread (raw->rowSize,  1, x, raw->file) != (UINT)x) {
			if (Verbose)
				printf ("%s: Could not read file correctly!\n",fileName);
			RawImageClose (raw);
			return NULL;
			}
		if (swapFlag) {
		    ConvertUint(raw->rowStart,(unsigned int) (x/sizeof(GLuint)));
		    ConvertUint((GLuint *)raw->rowSize,(unsigned int) (x/sizeof(GLint)));
			}
	    }
    return raw;
}

static void RawImageClose(rawImageRec *raw) {
    fclose(raw->file);
	if (raw -> tmp)
	    free(raw->tmp);
	if (raw -> tmpR)
	    free(raw->tmpR);
	if (raw -> tmpG)
	    free(raw->tmpG);
	if (raw -> tmpB)
	    free(raw->tmpB);
	if (raw)
	    free(raw);
}

bool RawImageGetRow (rawImageRec *raw, unsigned char *buf, int y, int z) {
    unsigned char *iPtr, *oPtr, pixel;
    int count;

    if ((raw->type & 0xFF00) == 0x0100) {
		fseek(raw->file, (long) raw->rowStart[y+z*raw->sizeY], SEEK_SET);
		if (fread (raw->tmp, 1, (unsigned int)raw->rowSize[y+z*raw->sizeY],raw->file) !=
			(unsigned int)raw->rowSize[y+z*raw->sizeY])
			return FALSE;

		iPtr = raw->tmp;
		oPtr = buf;
		for (;;) {
		    pixel = *iPtr++;
	    	count = (int)(pixel & 0x7F);
		    if (!count) 
				break;
		    if (pixel & 0x80) {
				while (count--)
		    		*oPtr++ = *iPtr++;
			    } 
			else {
				pixel = *iPtr++;
				while (count--)
				    *oPtr++ = pixel;
			    }
			}
	    } 
	else {
		fseek(raw->file, 512+(y*raw->sizeX)+(z*raw->sizeX*raw->sizeY),SEEK_SET);
		if (fread(buf, 1, raw->sizeX, raw->file) != raw->sizeX)
			return FALSE;
    	}
	return TRUE;
}

bool RawImageGetData (rawImageRec *raw, ImageRec *final) {
	bool Ret = TRUE;
    unsigned char *ptr;
    int i, j;

    final -> data = new unsigned char [(raw -> sizeX + 1) * (raw -> sizeY + 1) << 2];
    ptr = final->data;
    for (i = 0; i < (int)(raw->sizeY); i++) {
		if (!RawImageGetRow (raw, raw->tmpR, i, 0)) {
			Ret = FALSE;
			break;
			}
		if (!RawImageGetRow (raw, raw->tmpG, i, 1)) {
			Ret = FALSE;
			break;
			}
		if (!RawImageGetRow (raw, raw->tmpB, i, 2)) {
			Ret = FALSE;
			break;
			}
		for (j = 0; j < (int)(raw->sizeX); j++) {
	    	*ptr++ = *(raw->tmpR + j);
		    *ptr++ = *(raw->tmpG + j);
		    *ptr++ = *(raw->tmpB + j);
			}
	    }
	return Ret;
}

ImageRec *LoadRGB (const char *fileName,bool Verbose) {
    rawImageRec *raw;
    ImageRec *final = NULL;

    if ((raw = RawImageOpen (fileName,Verbose)) != NULL) {
	    final = new ImageRec;
    	final->sizeX = raw->sizeX;
	    final->sizeY = raw->sizeY;
		final->BPP   = 24;
    	if (!RawImageGetData (raw, final)) {
			if (Verbose)
				printf ("%s: Could not read file!\n",fileName);
			delete final;
			return NULL;
			}
	   	RawImageClose (raw);
		}
    return final;
}

/******************************************************************************/
//		BMP Funktionen
/******************************************************************************/
ImageRec *LoadBMP (const char *filename,bool Verbose) {
   	unsigned char *bits;
	unsigned char *Pal = NULL;
   	long bitsize;
   	BITMAPFILEHEADER header;
   	BITMAPINFO info;
	int fp;
	ImageRec *Im;
	
   	if ((fp = open (filename,OPEN_FLAGS)) == -1) {
		if (Verbose)
			printf("Could not open file for reading\n") ;
		}
   	else {
		read (fp,&header,sizeof (BITMAPFILEHEADER));
		if ((header.bfType != 0x424D) && (header.bfType != 0x4D42)) {
         	close (fp) ;
			if (Verbose)
	         	printf("Not a bitmap file!\n") ;
      		}
      	else if (read (fp,&info,sizeof (BITMAPINFOHEADER)) != sizeof (BITMAPINFOHEADER)) {
            close (fp) ;
			if (Verbose)
    	        printf ("Could not read the bitmap header\n") ;
        	}
        else {
			if (info.bmiHeader.biBitCount != 8 && info.bmiHeader.biBitCount != 24) {
				if (Verbose)
					printf ("Unsupported color depth!\n");
				return NULL;
				}
			if ((bitsize = info.bmiHeader.biSizeImage) == 0)
                  bitsize = info.bmiHeader.biWidth * (info.bmiHeader.biBitCount >> 3) *
                            info.bmiHeader.biHeight;
			if (info.bmiHeader.biBitCount == 8) {
				RGBQUAD Quad[256];
				int NumColor = info.bmiHeader.biClrUsed;
				if (read (fp,Quad,NumColor * sizeof (RGBQUAD)) != (int)(NumColor * sizeof (RGBQUAD))) {
    	            close (fp);
					if (Verbose)
        		        printf ("Could not read bitmap correctly!\n") ;
					return NULL;
            	    }
				int j = 0;
				Pal = new unsigned char [768];
				memset (Pal,0,768);
				for (int i = 0;i < NumColor;i++) {
					Pal[j++] = Quad[i].rgbRed;
					Pal[j++] = Quad[i].rgbGreen;
					Pal[j++] = Quad[i].rgbBlue;
					}
				}
            bits = new unsigned char [bitsize];
           	if (read (fp,bits,bitsize) != bitsize) {
				if (Pal)
					delete []Pal;
            	delete []bits;
                close (fp);
				if (Verbose)
        	        printf ("Could not read bitmap correctly!\n") ;
                }
            else {
                close (fp);
				Im = new ImageRec;
				Im -> sizeX = info.bmiHeader.biWidth;
				Im -> sizeY = info.bmiHeader.biHeight;
				Im -> data  = (unsigned char *)bits;
				Im -> BPP   = info.bmiHeader.biBitCount;
				Im -> Palette = Pal;
				if (Im -> BPP == 8)
					Im -> Convert8bitToRGB ();
				else
					Im -> SwapRB ();			// ???????????????
                return Im;
              	}
         	}
      	}
   	return (NULL) ;
}

/******************************************************************************/
//		TGA Funktionen
/******************************************************************************/
ImageRec *LoadTGA (const char *File,bool Verbose) {
	bool Compress = false;
	BYTE *Pal = NULL;
	char Dummy[256];
	int Fd;
	int ImageSize;
	ImageRec *Im;
	TGAHeader Head;
	
	if ((Fd = open (File,OPEN_FLAGS)) == -1) {
		if (Verbose)
			printf ("%s: Cannot open file for reading!\n", File);
		return NULL;
		}
	if (read (Fd,&Head,sizeof (TGAHeader)) != sizeof (TGAHeader)) {
		if (Verbose)
			printf ("%s: Could not read file correctly!\n",File);
		close (Fd);
		return NULL;
		}
	if (Head.PixelDepth != 24 && Head.PixelDepth != 32 && Head.PixelDepth != 8) {
		if (Verbose)
			printf ("Unsupported color depth!\n");
		close (Fd);
		return NULL;
		}
	if (Head.ImageType != 2 && Head.ImageType != 1 && 
		Head.ImageType != 9 && Head.ImageType != 10) {
		if (Verbose)
			printf ("Unsupported TGA type!\n");
		close (Fd);
		return NULL;
		}
	if (Head.ImageType == 9 || Head.ImageType == 10)
		Compress = true;
	// Read ID 	
	if (Head.IdLength) {
		if (read (Fd,Dummy,Head.IdLength) != Head.IdLength) {
			if (Verbose)
				printf ("%s: Could not read file correctly!\n",File);
			close (Fd);
			return NULL;
			}
		}
	if (Head.CmapType == 1) {	// Palette einlesen
		int Len = 3 * Head.CmapLength;
		Pal = new BYTE [Len];
		if (read (Fd,Pal,Len) != Len) {
			delete []Pal;
			close (Fd);
			if (Verbose)
				printf ("%s: Could not read file correctly!\n",File);
			return NULL;
			}
		}
	
	Im = new ImageRec;
	Im -> sizeX   = Head.ImageWidth;
	Im -> sizeY   = Head.ImageHeight;
	Im -> BPP     = Head.PixelDepth;
	Im -> Palette = Pal;
	ImageSize = Im -> sizeX * (Im -> BPP >> 3) * Im -> sizeY;

	Im -> data = new BYTE [ImageSize];
	if (!Compress) {
		if (read (Fd,Im -> data,ImageSize) != ImageSize) {
			if (Pal)
				delete []Pal;
			delete []Im -> data;
			delete Im;
			if (Verbose)
				printf ("%s: Could not read file correctly!\n",File);
			return NULL;
			}
		}
	else {
		int ReadBytes = GetFileSize (Fd);
		unsigned char *Temp = new unsigned char [ReadBytes];
		int i = 0,j;
		int Pos = 0;
		int Rep;
		int R,G,B,A;
		ReadBytes = read (Fd,Temp,ReadBytes);
		while (1) {
			if (Temp[i] & 0x80) {
				Rep = (Temp[i++] & 0x7F) + 1;
				switch (Im -> BPP) {
					case 8:
						R = Temp[i++];
						memset (Im -> data + Pos,R,Rep);
						Pos += Rep;
						break;
					case 24:
						R = Temp[i++];
						G = Temp[i++];
						B = Temp[i++];
						break;
					case 32:
						R = Temp[i++];
						G = Temp[i++];
						B = Temp[i++];
						A = Temp[i++];
						break;
					}
				for (j = 0;j < Rep;j++) {
					if (Im -> BPP == 8)
						break;
					switch (Im -> BPP) {
						case 24:
							Im -> data[Pos++] = R;
							Im -> data[Pos++] = G;
							Im -> data[Pos++] = B;
							break;
						case 32:
							Im -> data[Pos++] = R;
							Im -> data[Pos++] = G;
							Im -> data[Pos++] = B;
							Im -> data[Pos++] = A;
							break;
						}
					}
				}
			else {
				Rep = (Temp[i++] + 1) * (Im -> BPP >> 3);
				memmove (Im -> data + Pos,Temp + i,Rep);
				i += Rep;
				Pos += Rep;
				}
			if (Pos >= ImageSize || i >= ReadBytes)
				break;
			}
		delete []Temp;
		}
	if (Head.ImagDesc & 0x10) 
		Im -> HorizontalFlip ();
	if (!(Head.ImagDesc & 0x20))
		Im -> VerticalFlip ();
	if (Im -> BPP == 8)
		Im -> Convert8bitToRGB ();
	Im -> SwapRB ();			// ??????????????
	close (Fd);
	return Im;
}


////////////////////////////////////////////////////////////////////////////////////////
//
//	ImageRec Funktionen
//
////////////////////////////////////////////////////////////////////////////////////////
ImageRec::ImageRec () {
	Palette = NULL;
	data = NULL;
}

ImageRec::~ImageRec () {
	if (Palette)
		delete []Palette;
	if (data) 
		delete []data;
}

void ImageRec::Convert8bitToRGB () {
	if (BPP != 8)
		return;
		
	int SSize = sizeX * sizeY;
	int DSize = SSize * 3;
	int Pos = 0;
	unsigned char *NewData = new unsigned char [DSize];
	
	for (int i = 0;i < SSize;i++) {
		NewData[Pos++] = Palette[3 * data[i]];
		NewData[Pos++] = Palette[3 * data[i] + 1];
		NewData[Pos++] = Palette[3 * data[i] + 2];
		}
	BPP = 24;
	delete []data;
	data = NewData;
	delete []Palette;
	Palette = NULL;
}

void ImageRec::Convert8bitToRGBA (int Alpha) {
	if (BPP != 8)
		return;
		
	int SSize = sizeX * sizeY;
	int DSize = SSize << 2;
	int Pos = 0;
	unsigned char *NewData = new unsigned char [DSize];
	
	for (int i = 0;i < SSize;i++) {
		NewData[Pos++] = Palette[3 * data[i]];
		NewData[Pos++] = Palette[3 * data[i] + 1];
		NewData[Pos++] = Palette[3 * data[i] + 2];
		NewData[Pos++] = Alpha;
		}
	BPP = 32;
	delete []data;
	data = NewData;
	delete []Palette;
	Palette = NULL;
}

void ImageRec::ConvertRGB2RGBA (int Alpha) {
	if (BPP != 24)
		return;
		
	int SSize = sizeX * sizeY;
	int DSize = SSize << 2;
	int Pos = 0;
	int j = 0;
	unsigned char *NewData = new unsigned char [DSize];

	for (int i = 0;i < SSize;i++) {
		NewData[Pos++] = data[j++];
		NewData[Pos++] = data[j++];
		NewData[Pos++] = data[j++];
		NewData[Pos++] = Alpha;
		}
	BPP = 32;
	delete []data;
	data = NewData;
}

void ImageRec::SetAlpha (int Alpha) {
	if (BPP != 32)
		return;
		
	int Size = sizeX * sizeY << 2;
	for (int i = 3;i < Size;i += 4)
		data[i] = Alpha;
}

bool ImageRec::HorizontalFlip () {
	int i;
	int Bytes = BPP >> 3;
	int Line = 0;
	int LineLeft = 0;
	int ScanLine = sizeX * Bytes;
	int LineRight = ScanLine - Bytes;
	int HalfLine = ScanLine / Bytes >> 1;
	BYTE R,G,B;
	BYTE Pixel[4];
	
	if (BPP != 24 && BPP != 32 && BPP != 8) {
		printf ("Unsupported operation for this color depth!\n");
		return FALSE;
		}
	for (int j = 0;j < sizeY;j++) {
		for (i = 0;i < HalfLine;i++) {
			switch (BPP) {
				case 8:
					R = data[LineLeft];
					data[LineLeft] = data[LineRight];
					data[LineRight] = R;
					break;
				case 24:
					R = data[LineLeft];
					G = data[LineLeft + 1];
					B = data[LineLeft + 2];
					data[LineLeft]     = data[LineRight];
					data[LineLeft + 1] = data[LineRight + 1];
					data[LineLeft + 2] = data[LineRight + 2];
					data[LineRight]     = R;
					data[LineRight + 1] = G;
					data[LineRight + 2] = B;
					break;
				case 32:
					memmove (Pixel,data + LineLeft,Bytes);
					memmove (data + LineLeft,data + LineRight,Bytes);
					memmove (data + LineRight,Pixel,Bytes);
					break;
				}
			LineLeft  += Bytes;
			LineRight -= Bytes;
			}
		Line += ScanLine;
		LineLeft  = Line;
		LineRight = Line + ScanLine - Bytes;
		}
	return TRUE;
}

bool ImageRec::VerticalFlip () {
	int ScanLine = sizeX * (BPP >> 3);
	BYTE *Dummy = new BYTE [ScanLine];
	int ImageSize = ScanLine * sizeY;
	int LineTop = 0;
	int LineBottom = ImageSize - ScanLine;
	for (int j = 0;j < sizeY;j++) {
		memmove (Dummy,data + LineBottom,ScanLine);
		memmove (data + LineBottom,data + LineTop,ScanLine);
		memmove (data + LineTop,Dummy,ScanLine);
		}
	delete []Dummy;
	return TRUE;
}

bool ImageRec::SwapRB () {
	BYTE R;
	int Pos = 0;
	int Add = BPP >> 3;
	int i;

	if (BPP != 24 && BPP != 32)
		return FALSE;	
	for (int j = 0;j < sizeY;j++) {
		for (i = 0;i < sizeX;i++) {
			R = data[Pos];
			data[Pos] = data[Pos + 2];
			data[Pos + 2] = R;
			Pos += Add;
			}
		}
	return TRUE;
}
